'use strict';

// ---------------------------------------------------------
// "Classes" that represent AST nodes
// ---------------------------------------------------------

class Program extends AST {
  constructor(ss) {
    super();
    this.ss = ss;
  }
  translate(data) {
	for (var i = 0; i < this.ss.length; ++i) {
		if (i === this.ss.length-1) {
			data.lastFlag = "print";
		}
		this.ss[i].translate(data);
	}
	data.lastFlag = "";
  }
}

// Statements

class Stmt extends AST {
  constructor() {
    super();
  }
  translate(data) {}
}

class ClassDecl extends Stmt {
  constructor(C, S, xs) {
    super();
    this.C = C;
    this.S = S;
    this.xs = xs;
	this.methodNames = new Set();
  }
  declareIt(x) {
	if (this.first) {
		this.first = false;
	  	return "Obj " + x;
	} else {
		return ", Obj " + x;
	}
  }
  translate(data) {
	// add a constructor declaration
	var conDecl = "Obj " + this.C + "_new(";
	this.first = true;
	this.xs.forEach(x => (conDecl += this.declareIt(x)));
	conDecl += ")";
	this.conDecl = conDecl;
    data.conDecls += conDecl + ";\n"
	
	// add myself to the global map of classes
	data.classes.set(this.C, this);
	  
	// add prettyPrint function
	var prettyPrintMethod = "Obj " + this.C + "_prettyPrint(Obj this) {\n";
	prettyPrintMethod += '    printf("' + this.C + '(");\n';
	this.xs.forEach(x => (prettyPrintMethod += '    ((Obj (*)(Obj))this->'+x+'->prettyPrint)(this->'+x+');\n    printf(",");\n'));
	prettyPrintMethod += '    printf(")");\n';
	prettyPrintMethod += "}\n";
	data.methods += prettyPrintMethod;
	  
	// update data.fieldNames
	this.xs.forEach(x => data.fieldNames.add(x));
  }

  addConstructor(data) {
	var conDefn = this.conDecl + " {\n";
	conDefn += "    Obj result = malloc(sizeof(struct Object));\n";
	conDefn += "    result->classname = \"" + this.C + "\";\n";
	conDefn += "    result->prettyPrint = &" + this.C + "_prettyPrint;\n";
		  
    // bind methods
	this.methodNames.forEach(m => (conDefn += "result->" + m + " = &" + this.C + "_" + m + ";\n"));

    // initialize the object by calling init
    var paramTypeList = "Obj";
	this.xs.forEach(x => (paramTypeList += ", Obj"));
    var paramList = "result";
	this.xs.forEach(x => (paramList += ", " + x));
	conDefn += "    ((Obj (*)(" + paramTypeList + "))result->init)("+paramList+");\n";
	
	// return the result
	conDefn += "    return result;\n";
	conDefn += "}\n";
	  
	// add it to the constructor definition string
    data.cons += conDefn;
  }
}

class MethodDecl extends Stmt {
  constructor(C, m, xs, ss) {
    super();
    this.C = C;
    this.m = m;
    this.xs = xs;
    this.ss = ss;
  }
  translate(data) {
	  data.methodNames.add(this.m);
	  data.classes.get(this.C).methodNames.add(this.m);
	  
	  // declare the method
	  var args = "";
	  this.xs.forEach(x => (args += ", Obj " + x));
	  var method = "Obj " + this.C + "_" + this.m + "(Obj this" + args + ") {\n";

	  // save the outer body
	  const outerBody = data.currentBody;
	  data.currentBody = "";
	  
	  // output statements
	  for (var i = 0; i < this.ss.length; ++i) {
		  if (i === this.ss.length-1) {
		  	  data.lastFlag = "return";
		  }
		  this.ss[i].translate(data);
	  }
	  data.lastFlag = "";
	  
	  method += data.currentBody;
	  method += "}\n";
	  data.methods += method;
	  // restore the outer body
	  data.currentBody = outerBody;
  }
}

class VarDecl extends Stmt {
  constructor(x, e) {
    super();
    this.x = x;
    this.e = e;
  }
  translate(data) {
	  data.currentBody += "Obj " + this.x + " = ";
	  this.e.translate(data);
	  data.currentBody += ";\n";
  }
}

class VarAssign extends Stmt {
  constructor(x, e) {
    super();
    this.x = x;
    this.e = e;
  }
  translate(data) {
	  data.currentBody += this.x + " = ";
	  this.e.translate(data);
	  data.currentBody += ";\n";
  }
}

class InstVarAssign extends Stmt {
  constructor(x, e) {
    super();
    this.x = x;
    this.e = e;
  }
  translate(data) {
	  data.currentBody += "this->" + this.x + " = ";
	  this.e.translate(data);
	  data.currentBody += ";\n";
  }
}

class Return extends Stmt {
  constructor(e) {
    super();
    this.e = e;
  }
}

class ExpStmt extends Stmt {
  constructor(e) {
    super();
    this.e = e;
  }
  translate(data) {
	  if (data.lastFlag === "return") {
	  	  data.currentBody += "return ";
	  }
	  if (data.lastFlag === "print") {
	  	  data.currentBody += 'Obj the_answer = ';
	  }
	  this.e.translate(data);
	  if (data.lastFlag === "print") {
	  	  data.currentBody += ';\n((Obj (*)(Obj))the_answer->prettyPrint)(the_answer)';
	  }
	  data.currentBody += ";\n";
  }
}

// Expressions

class Exp extends AST {
  constructor() {
    super();
  }
  translate(data) {
	  data.currentBody += "0";
  }
}

class Lit extends Exp {
  constructor(primValue) {
    super();
    this.primValue = primValue;
  }
}

class Var extends Exp {
  constructor(x) {
    super();
    this.x = x;
  }
  translate(data) {
	  data.currentBody += this.x;
  }
}

class BinOp extends Exp {
  constructor(op, e1, e2) {
    super();
    this.op = op;
    this.e1 = e1;
    this.e2 = e2;
  }
}

class This extends Exp {
  constructor() {
    super();
  }
  translate(data) {
	  data.currentBody += "this";
  }
}

class InstVar extends Exp {
  constructor(x) {
    super();
    this.x = x;
  }
  translate(data) {
	  data.currentBody += "this->" + this.x;
  }
}

class New extends Exp {
  constructor(C, es) {
    super();
    this.C = C;
    this.es = es;
  }
  passIt(data, e) {
	if (this.first) {
		this.first = false;
	} else {
		data.currentBody += ", ";
	}
	e.translate(data);
  }
  translate(data) {
	  var args = "";
	  this.first = true;
	  data.currentBody += this.C + "_new(";
	  this.es.forEach(e => (this.passIt(data, e)));
	  data.currentBody += ")";
  }
}

class Send extends Exp {
  constructor(erecv, m, es) {
    super();
    this.erecv = erecv;
    this.m = m;
    this.es = es;
  }
  passIt(data, e) {
	if (this.first) {
		this.first = false;
	} else {
		data.currentBody += ", ";
	}
	e.translate(data);
  }
  translate(data) {
      var paramTypeList = "Obj";
	  this.es.forEach(x => (paramTypeList += ", Obj"));
	  data.currentBody += "((Obj (*)(" + paramTypeList + "))";
	  this.erecv.translate(data);
	  data.currentBody += "->" + this.m + ")(";
	  this.erecv.translate(data);
	  var args = "";
	  this.first = false;
	  this.es.forEach(e => (this.passIt(data, e)));
	  data.currentBody += ")";
  }
}

class SuperSend extends Exp {
  constructor(m, es) {
    super();
    this.m = m;
    this.es = es;
  }
}

class BlockLit extends AST {
  constructor(xs, ss) {
    super();
    this.xs = xs;
    this.ss = ss;
  }
}

